==> quadrant.transit-0.8.12.tgz: package/src/lib/Command.ts <==
import { stringify } from "querystring";

/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

export type EntityType = 'table' | 'view' | 'procedure' | 'function';

export enum DataType {
	Null      = 0,
	Boolean   = 1,
	Integer   = 2,
	Float     = 3,
	String    = 4,
	Date      = 5,
	DateTime  = 6,
	Array     = 7,
	Object    = 8,
	Any       = 9,
}

export interface EntitySpec {
	readonly name: string;
	readonly type: EntityType;
	readonly identity: FieldSpec[];
	readonly impl: any;
	readonly operations: { [key: string] : OperationSpec };
}

export interface OperationSpec {
	readonly name: string;
	readonly parameters: ParameterSpec[];
	readonly outputs: OutputSpec[];
}

export interface FieldSpec {
    readonly name: string;
	readonly dataType: DataType;
	readonly implType?: string;
}

export interface ParameterSpec {
	readonly spec: FieldSpec;
	readonly required: boolean;
	readonly subparams: ParameterSpec[]; // for type Object, if this is empty the object members will go unchecked...
}

export interface OutputSpec {
	readonly spec: FieldSpec;
	readonly writeable: boolean;
}

export interface CommandSpec {
	readonly entity : EntitySpec;
	readonly operation : OperationSpec;
}

export type EntityNameEscaper = (entityName : string) => string;
export type EntityNamespacer = (entity : EntitySpec) => string;

export class Command {
	readonly spec : CommandSpec;
	readonly args : any;

	constructor(spec: CommandSpec, args : any[]) {
		this.spec = spec;
		this.args = spec.operation.parameters.map((param, i) => {
			let arg = this.validateArgument(param, args[i]);
			if (arg === undefined) {
				throw new Error(`Parameter ${param.spec.name} was not specified`);
			} else {
				return arg;
			}
		});
	}

	validateArgument(param : ParameterSpec, inarg : any) : any {
		var error = (type) => {
			throw new Error(`Argument "${inarg}" for parameter "${param.spec.name}" of entity "${this.spec.entity.name}" is not of data type "${type}".`);
		}

		if (inarg == null && param.required) {
			throw new Error(`Parameter ${param.spec.name} of ${this.spec.entity.name}.${this.spec.operation.name} is required.`);
		}

		// validate data types
		if ((inarg != undefined && inarg != null) || param.required) {
			let num, d;
			switch (param.spec.dataType) {
				case DataType.Null:    
					if (typeof inarg != "string" && inarg != "" && inarg.toLowerCase() != "null") error("null"); 
					break;
				case DataType.Boolean:
					if (typeof inarg != "boolean") error("boolean");
					break;
				case DataType.Integer:
					if (typeof inarg != "number" || !Number.isInteger(inarg)) error("integer");
					break;
				case DataType.Float:
					if (typeof inarg != "number" || isNaN(inarg)) error("float");
					break;
				case DataType.String:
					if (typeof inarg != "string") error("string");
					break;
				case DataType.Date:
					num = Date.parse(inarg);
					d = new Date(num);
					if (isNaN(num) || d.getUTCHours() != 0 || d.getUTCMinutes() != 0 || d.getUTCSeconds() != 0 || d.getUTCMilliseconds() != 0) error("date");
					break;
				case DataType.DateTime:
					num = Date.parse(inarg);
					if (typeof inarg != "string" || (isNaN(num) && !inarg.match(/^\d\d:\d\d:\d\d$/)) ) error("date/time");
					break;
				case DataType.Array:
					if (inarg == null) inarg = [];
					if (typeof inarg != "object" || !Array.isArray(inarg)) error("array");
					break;
				case DataType.Object:
					if (typeof inarg != "object" || Array.isArray(inarg)) error("object");
					break;
				case DataType.Any:
					break;
				default:
					throw new Error(`internal error: invalid data type: ${param.spec.dataType}`);
			}
		}

		// now translate defaults, if necessary
		// be careful, inarg may be undefined...
		let outarg = inarg;
		switch (param.spec.dataType) {
			case DataType.Null:    
				outarg = null;
				break;
			case DataType.Boolean:
			case DataType.Integer:
			case DataType.Float:
			case DataType.String:
			case DataType.Date:
			case DataType.DateTime:
				break;
			case DataType.Array:
				if (inarg == undefined) inarg = [];
				outarg = [];
				for (let element of inarg) {
					let elemArg = this.validateArgument(param.subparams[0], element);
					if (elemArg == null) {
						// can this even happen?
						throw new Error("Array element value was not specified");
					}
					outarg.push(elemArg);
				}
				break;
			case DataType.Object:
				if (param.subparams.length == 0) {
					outarg = inarg || {};
				} else {
					outarg = {};
					for (let subparam of param.subparams) {
						let subarg = this.validateArgument(subparam, inarg[subparam.spec.name]);
						if (subarg !== undefined) {
							outarg[subparam.spec.name] = subarg;
						}
					}
					for (let name in inarg) {
						if (!(name in outarg) && inarg[name] !== undefined) {
							throw new Error(`Unknown object key "${name}" for parameter "${param.spec.name}" of entity "${this.spec.entity.name}"`);
						}
					}
				}
				break;
			case DataType.Any:
				if (outarg === undefined) outarg = null;
				break;
			default:
				throw new Error(`internal error: invalid data type: ${param.spec.dataType}`);
		}

		return outarg;
	}

	static normalizeName(name : string, env : Object) {
		name = name.replace(/^[^\w-$]+/, "").replace(/[^\w-$]+$/, "").replace(/[^\w-$]+/g, "_");
		if (env[name]) {
			var i = 2;
			while (env[name + i]) i++;
			name += i;
		}
		return name;
	}
}

export class EntityBuilder {
	private resolve:Function;
	private reject:Function;
	readonly promise:Promise<EntitySpec[]> = new Promise<EntitySpec[]>((resolve, reject) => {
		this.resolve = resolve;
		this.reject = reject;
	});

	protected entities:{[key:string]:EntitySpec} = {};
	protected currentEntity:EntitySpec;
	protected currentOperation:OperationSpec;
	protected currentParameter:ParameterSpec;

	entity(name:string, type: EntityType, impl:any):EntityBuilder {
		this.entities[name] = this.currentEntity = this.entities[name] || {
			name: name,
			type: type,
			impl: impl,
			identity: [],
			operations: {}
		}
		return this;
	}

	operation(name:string):EntityBuilder {
		this.currentEntity.operations[name] = this.currentOperation = this.currentEntity.operations[name] || {
			name: name,
			parameters: [],
			outputs: []
		};
		return this;
	}

	identity(name:string, dataType:DataType, implType:string) {
		this.currentEntity.identity.push({
			name: name,
			dataType: dataType,
			implType: implType
		});
	}

	parameter(name:string, dataType:DataType, implType:string|undefined, required:boolean):EntityBuilder {
		if (!this.currentOperation.parameters.find((op) => op.spec.name == name)) {
			this.currentOperation.parameters.push(this.currentParameter = {
				spec: {
					name: name,
					dataType: dataType,
					implType: implType
				},
				required: required,
				subparams: []
			});
		}
		this.currentParameter = this.currentOperation.parameters.find((op) => op.spec.name == name)!;
		return this;
	}

	subparam(name:string, dataType:DataType, implType:string|undefined, required:boolean):EntityBuilder {
		if (!this.currentParameter.subparams.find((param) => param.spec.name == name)) {
			this.currentParameter.subparams.push({
				spec: {
					name: name,
					dataType: dataType,
					implType: implType
				},
				required: required,
				subparams: []
			});
		}
		return this;
	}

	output(name:string, dataType:DataType, implType:string|undefined, writeable:boolean):EntityBuilder {
		this.currentOperation.outputs.push({
			spec: {
				name: name,
				dataType: dataType,
				implType: implType
			},
			writeable: writeable
		});
		return this;
	}

	done() {
		this.resolve(Object["values"](this.entities));
	}
}

export class DatabaseEntityBuilder extends EntityBuilder {
	// for procs that take no params and return nothing
	proc() {
		this.operation("call");
	}

	procParameter(name:string, dataType:DataType, implType:string, required:boolean) {
		this.operation("call").parameter(name, dataType, implType, required);
	}

	procOutput(name:string, dataType:DataType, implType:string) {
		this.operation("call").output(name, dataType, implType, false);
	}

	tableColumn(name:string, dataType:DataType, implType:string, required:boolean, primaryKey:boolean, writeable:boolean):EntityBuilder {
		this.column(name, dataType, implType, writeable);
		this.operation("truncate");
		if (primaryKey) {
			this.identity(name, dataType, implType);
			this.operation("insert").parameter(name, dataType, implType, writeable);
			this.operation("update").parameter(name, dataType, implType, true);
			this.operation("delete").parameter(name, dataType, implType, true);
		} else {
			this.operation("insert")
				.parameter("values", DataType.Object, "object", true)
					.subparam(name, dataType, implType, required && writeable);
			
			// we can insert w/o primary key, but we can't update
			if (this.currentEntity.identity.length) {
				if (this.currentEntity.identity.length == this.currentEntity.operations.update.parameters.length) {
					// it's possible to change pks on update, so add them as values...
					this.currentEntity.identity.forEach((pk) => {
						this.operation("update").parameter("values", DataType.Object, "object", true)
							.subparam(pk.name, pk.dataType, pk.implType, false);
					});
				}


				this.operation("update")
					.parameter("values", DataType.Object, "object", true)
						.subparam(name, dataType, implType, false);
			}
		}
		return this;
	}

	viewColumn(name:string, dataType:DataType, implType:string):EntityBuilder {
		this.column(name, dataType, implType, false);
		return this;
	}

	column(name:string, dataType:DataType, implType:string, writeable:boolean):EntityBuilder {
		this.operation("select")
			.parameter("columns", DataType.Array, "array", false)
				.subparam("column", DataType.String, "string", false)
			.parameter("filter", DataType.Any, "any", false)
			.parameter("sort", DataType.Array, "array", false)
				.subparam("columns", DataType.String, "string", false);
		this.operation("select").output(name, dataType, implType, writeable);
		return this;
	}
}

==> quadrant.transit-0.8.17.tgz: package/src/lib/Command.ts <==
import { stringify } from "querystring";

/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only. 
 */

export type EntityType = 'table' | 'view' | 'procedure' | 'function';

export enum DataType {
	Null      = 0,
	Boolean   = 1,
	Integer   = 2,
	Float     = 3,
	String    = 4,
	Date      = 5,
	DateTime  = 6,
	Array     = 7,
	Object    = 8,
	Any       = 9,
}

export interface EntitySpec {
	readonly name: string;
	readonly type: EntityType;
	readonly identity: FieldSpec[];
	readonly impl: any;
	readonly operations: { [key: string] : OperationSpec };
}

export interface OperationSpec {
	readonly name: string;
	readonly parameters: ParameterSpec[];
	readonly outputs: OutputSpec[];
}

export interface FieldSpec {
    readonly name: string;
	readonly dataType: DataType;
	readonly implType?: string;
}

export interface ParameterSpec {
	readonly spec: FieldSpec;
	readonly required: boolean;
	readonly subparams: ParameterSpec[]; // for type Object, if this is empty the object members will go unchecked...
}

export interface OutputSpec {
	readonly spec: FieldSpec;
	readonly writeable: boolean;
}

export interface CommandSpec {
	readonly entity : EntitySpec;
	readonly operation : OperationSpec;
}

export type EntityNameEscaper = (entityName : string) => string;
export type EntityNamespacer = (entity : EntitySpec) => string;

export class Command {
	readonly spec : CommandSpec;
	readonly args : any;

	constructor(spec: CommandSpec, args : any[]) {
		this.spec = spec;
		this.args = spec.operation.parameters.map((param, i) => {
			let arg = this.validateArgument(param, args[i]);
			if (arg === undefined) {
				throw new Error(`Parameter ${param.spec.name} was not specified`);
			} else {
				return arg;
			}
		});
	}

	validateArgument(param : ParameterSpec, inarg : any) : any {
		var error = (type) => {
			throw new Error(`Argument "${inarg}" for parameter "${param.spec.name}" of entity "${this.spec.entity.name}" is not of data type "${type}".`);
		}

		if (inarg == null && param.required) {
			throw new Error(`Parameter ${param.spec.name} of ${this.spec.entity.name}.${this.spec.operation.name} is required.`);
		}

		// validate data types
		if ((inarg != undefined && inarg != null) || param.required) {
			let num, d;
			switch (param.spec.dataType) {
				case DataType.Null:    
					if (typeof inarg != "string" && inarg != "" && inarg.toLowerCase() != "null") error("null"); 
					break;
				case DataType.Boolean:
					if (typeof inarg != "boolean") error("boolean");
					break;
				case DataType.Integer:
					if (typeof inarg != "number" || !Number.isInteger(inarg)) error("integer");
					break;
				case DataType.Float:
					if (typeof inarg != "number" || isNaN(inarg)) error("float");
					break;
				case DataType.String:
					if (typeof inarg != "string") error("string");
					break;
				case DataType.Date:
					num = Date.parse(inarg);
					d = new Date(num);
					if (isNaN(num) || d.getUTCHours() != 0 || d.getUTCMinutes() != 0 || d.getUTCSeconds() != 0 || d.getUTCMilliseconds() != 0) error("date");
					break;
				case DataType.DateTime:
					num = Date.parse(inarg);
					if (typeof inarg != "string" || (isNaN(num) && !inarg.match(/^\d\d:\d\d:\d\d$/)) ) error("date/time");
					break;
				case DataType.Array:
					if (inarg == null) inarg = [];
					if (typeof inarg != "object" || !Array.isArray(inarg)) error("array");
					break;
				case DataType.Object:
					if (typeof inarg != "object" || Array.isArray(inarg)) error("object");
					break;
				case DataType.Any:
					break;
				default:
					throw new Error(`internal error: invalid data type: ${param.spec.dataType}`);
			}
		}

		// now translate defaults, if necessary
		// be careful, inarg may be undefined...
		let outarg = inarg;
		switch (param.spec.dataType) {
			case DataType.Null:    
				outarg = null;
				break;
			case DataType.Boolean:
			case DataType.Integer:
			case DataType.Float:
			case DataType.String:
			case DataType.Date:
			case DataType.DateTime:
				break;
			case DataType.Array:
				if (inarg == undefined) inarg = [];
				outarg = [];
				for (let element of inarg) {
					let elemArg = this.validateArgument(param.subparams[0], element);
					if (elemArg == null) {
						// can this even happen?
						throw new Error("Array element value was not specified");
					}
					outarg.push(elemArg);
				}
				break;
			case DataType.Object:
				if (!inarg) inarg = {};
				if (param.subparams.length == 0) {
					outarg = inarg;
				} else {
					outarg = {};
					for (let subparam of param.subparams) {
						let subarg = this.validateArgument(subparam, inarg[subparam.spec.name]);
						if (subarg !== undefined) {
							outarg[subparam.spec.name] = subarg;
						}
					}
					for (let name in inarg) {
						if (!(name in outarg) && inarg[name] !== undefined) {
							throw new Error(`Unknown object key "${name}" for parameter "${param.spec.name}" of entity "${this.spec.entity.name}"`);
						}
					}
				}
				break;
			case DataType.Any:
				if (outarg === undefined) outarg = null;
				break;
			default:
				throw new Error(`internal error: invalid data type: ${param.spec.dataType}`);
		}

		return outarg;
	}

	static normalizeName(name : string, env : Object) {
		name = name.replace(/^[^\w-$]+/, "").replace(/[^\w-$]+$/, "").replace(/[^\w-$]+/g, "_");
		if (env[name]) {
			var i = 2;
			while (env[name + i]) i++;
			name += i;
		}
		return name;
	}
}

export class EntityBuilder {
	private resolve:Function;
	private reject:Function;
	readonly promise:Promise<EntitySpec[]> = new Promise<EntitySpec[]>((resolve, reject) => {
		this.resolve = resolve;
		this.reject = reject;
	});

	protected entities:{[key:string]:EntitySpec} = {};
	protected currentEntity:EntitySpec;
	protected currentOperation:OperationSpec;
	protected currentParameter:ParameterSpec;

	entity(name:string, type: EntityType, impl:any):EntityBuilder {
		this.entities[name] = this.currentEntity = this.entities[name] || {
			name: name,
			type: type,
			impl: impl,
			identity: [],
			operations: {}
		}
		return this;
	}

	operation(name:string):EntityBuilder {
		this.currentEntity.operations[name] = this.currentOperation = this.currentEntity.operations[name] || {
			name: name,
			parameters: [],
			outputs: []
		};
		return this;
	}

	identity(name:string, dataType:DataType, implType:string) {
		this.currentEntity.identity.push({
			name: name,
			dataType: dataType,
			implType: implType
		});
	}

	parameter(name:string, dataType:DataType, implType:string|undefined, required:boolean):EntityBuilder {
		if (!this.currentOperation.parameters.find((op) => op.spec.name == name)) {
			this.currentOperation.parameters.push(this.currentParameter = {
				spec: {
					name: name,
					dataType: dataType,
					implType: implType
				},
				required: required,
				subparams: []
			});
		}
		this.currentParameter = this.currentOperation.parameters.find((op) => op.spec.name == name)!;
		return this;
	}

	subparam(name:string, dataType:DataType, implType:string|undefined, required:boolean):EntityBuilder {
		if (!this.currentParameter.subparams.find((param) => param.spec.name == name)) {
			this.currentParameter.subparams.push({
				spec: {
					name: name,
					dataType: dataType,
					implType: implType
				},
				required: required,
				subparams: []
			});
		}
		return this;
	}

	output(name:string, dataType:DataType, implType:string|undefined, writeable:boolean):EntityBuilder {
		this.currentOperation.outputs.push({
			spec: {
				name: name,
				dataType: dataType,
				implType: implType
			},
			writeable: writeable
		});
		return this;
	}

	outputOne(name:string, dataType:DataType) {
		if (!this.currentOperation.outputs.length) {
			this.output(name, dataType, undefined, false);
		}
	}

	done() {
		this.resolve(Object["values"](this.entities));
	}
}

export class DatabaseEntityBuilder extends EntityBuilder {
	// for procs that take no params and return nothing
	proc() {
		this.operation("call");
	}

	procParameter(name:string, dataType:DataType, implType:string, required:boolean) {
		this.operation("call").parameter(name, dataType, implType, required);
	}

	procOutput(name:string, dataType:DataType, implType:string) {
		this.operation("call").output(name, dataType, implType, false);
	}

	tableColumn(name:string, dataType:DataType, implType:string, required:boolean, primaryKey:boolean, writeable:boolean):EntityBuilder {
		this.column(name, dataType, implType, writeable);
		this.operation("truncate");
		if (primaryKey) {
			this.identity(name, dataType, implType);
			this.operation("insert").parameter(name, dataType, implType, writeable);
			this.operation("update").parameter(name, dataType, implType, true);
			this.operation("delete").parameter(name, dataType, implType, true);
		} else {
			this.operation("insert")
				.parameter("values", DataType.Object, "object", true)
					.subparam(name, dataType, implType, required && writeable);
			
			// we can insert w/o primary key, but we can't update
			if (this.currentEntity.identity.length) {
				if (this.currentEntity.identity.length == this.currentEntity.operations.update.parameters.length) {
					// it's possible to change pks on update, so add them as values...
					this.currentEntity.identity.forEach((pk) => {
						this.operation("update").parameter("values", DataType.Object, "object", true)
							.subparam(pk.name, pk.dataType, pk.implType, false);
					});
				}


				this.operation("update")
					.parameter("values", DataType.Object, "object", true)
						.subparam(name, dataType, implType, false);
			}
		}
		return this;
	}

	viewColumn(name:string, dataType:DataType, implType:string):EntityBuilder {
		this.column(name, dataType, implType, false);
		return this;
	}

	column(name:string, dataType:DataType, implType:string, writeable:boolean):EntityBuilder {
		this.operation("select")
			.parameter("columns", DataType.Array, "array", false)
				.subparam("column", DataType.String, "string", false)
			.parameter("filter", DataType.Any, "any", false)
			.parameter("sort", DataType.Array, "array", false)
				.subparam("columns", DataType.String, "string", false)
			.parameter("options", DataType.Object, "object", false)
				.subparam("pageSize", DataType.Integer, undefined, false)
				.subparam("pageIndex", DataType.Integer, undefined, false);
		this.operation("select").output(name, dataType, implType, writeable);

		this.operation("count")
			.parameter("filter", DataType.Any, "any", false);
		this.operation("count").outputOne("count", DataType.Integer);
		return this;
	}
}
